from copy import deepcopy
from typing import Any
from hhat_lang.interpreter.memory import Var, R
from hhat_lang.syntax_trees.ast import AOT
from hhat_lang.datatypes.builtin_datatype import (
    builtin_data_types_dict,
    builtin_array_types_dict,
    DefaultType
)
from hhat_lang.datatypes.base_datatype import DataType, DataTypeArray
from hhat_lang.interpreter.builtins.builtin_fn import builtin_fn_dict
from hhat_lang.interpreter.memory import Mem


class Eval:
    def __init__(self, code: R):
        self.code = code

    def run(self):
        mem = Mem()
        execute(self.code, mem)
        print(mem)


def eval_token(code: AOT, mem: Mem) -> Any:
    if code.type in ["oper", "id"]:
        # print(f"  = oper:", end=" ")
        if res := builtin_fn_dict.get(code.token, False):
            return res
        if code.token in mem:
            return mem.get_var(code.token)
        return Var(code.token)
    # print(f"  = literal:", end=" ")
    return builtin_data_types_dict.get(code.type, DefaultType)(code.token)


def eval_oper(code: R, mem: Mem) -> Any:
    # print("* oper:")
    res = ()
    # print(f"  -> oper content: {code} {[(k, type(k)) for k in code]}")
    for k in code:
        last = execute(k, mem)
        if isinstance(last[0], Var):
            # print(f"  => {type(k)} {code.role}")
            if code.role == "callee":
                # print("  ! oper callee found!")
                var = mem.get_var(last[0].name)
                # print(f"  ! [oper callee] after mem: {mem}")
                res += var,
            else:
                # print("IS VAR!!")
                mem.put_expr(last[0])
                res += last
        elif isinstance(last[0], (DataType, DataTypeArray)):
            mem.put_stack(last[0])
            res += last
        else:
            oper = last[0](mem, *mem.get_stack())
            # print(f"---> {type(oper)} {oper}")
            mem.put_expr(oper)
            # mem.put_stack(oper)
            res += oper,
    # print(f"  -> oper res: {res}")
    return res


def eval_args(code: R, mem: Mem) -> Any:
    # print("* args:")
    res = ()
    # print(f"    -> {len(code)} {code}")
    for k in code:
        # print(f"  !=> what is k: {k} | {mem}")
        last = execute(k, mem)
        # print(f"  => arg data: {last[0]} ({type(last[0])}) {type(last)} ({code}) {mem=}")
        mem.put_stack(last[0])
        res += last
    return res


def eval_call(code: R, mem: Mem) -> Any:
    # print("* call:")
    res = ()
    for k in code:
        res += execute(k, mem)
    # print(f"! what is call res: {res}")
    if len(code) == 2:
        args = mem.pop_stack()
        oper = mem.pop_expr()
        # print(f"call: {oper=} | {args=} ({type(args)})")
        new_res = oper(args)
        for p in new_res:
            mem.put_stack(p)
        new_res = ()
    else:
        if isinstance(res[0], Var):
            if res[0].initialized:
                new_res = res
            else:
                mem.pop_expr()
                # print(f"VAR!? {code=} | {res[0]=} | {mem=}")
                new_res = res[0](mem.pop_stack()),
                mem.put_var(res[0], "")
                mem.put_stack(res[0])
        else:
            oper = mem.pop_expr()
            if oper.token in builtin_fn_dict.keys():
                new_res = oper()
                # print(f"* [{oper}] received {new_res=}")
                for p in new_res:
                    mem.put_stack(p)
            else:
                print("WHAT")
                new_res = res
    return new_res


def eval_array(code: R, mem: Mem) -> Any:
    # print("* array:")
    res = ()
    for k in code:
        res += execute(k, mem)
    # print(f"! {res=} {len(set(k.type for k in res))=}")
    if len(set(k.type for k in res)) == 1:
        if isinstance(res[0], DataType):
            array = builtin_array_types_dict[res[0].type](*res)
            mem.put_stack(array)
            res = array,
        elif isinstance(res[0], int):
            print("* WE GOT AN INT!!")
    return res


def eval_expr(code: R, mem: Mem) -> Any:
    # print("* expr:")
    res = ()
    for k in code:
        res += execute(k, mem)
        # print(f"  [end][expr]-> after mem: {mem} | {res=}")
    return (res[-1],) if res else ()


def eval_many_expr(code: R, mem: Mem) -> Any:
    # print("* many-expr:")
    # print(f"  [cur]-> {mem}")
    res = ()
    for n, k in enumerate(code):
        new_mem = deepcopy(mem)
        # print(f"  [{n}][start][many-expr]({k})")
        # print(f"     -> prev mem: {mem}")
        # print(f"     -> new mem: {new_mem}")
        res += execute(k, new_mem)
        new_mem.share_vars(mem)
        # print(f"  [{n}][end][many-expr]({k}) -> after new mem: {new_mem} | mem: {mem} | {res=}")
    mem.clear_stack()
    for k in res:
        mem.put_stack(k)
    return res


def eval_main(code: R, mem: Mem) -> Any:
    # print("* main:")
    res = ()
    for k in code:
        res += execute(k, mem),
        mem.clear_stack()
    return res


def execute(code: Any, mem: Mem) -> tuple[Any]:
    res = ()
    match code:
        case R():
            match code.type:
                case "program":
                    pass

                case "main":
                    res = eval_main(code, mem)

                case "many-expr":
                    res = eval_many_expr(code, mem)

                case "expr":
                    res = eval_expr(code, mem)

                case "array":
                    res = eval_array(code, mem)

                case "call":
                    res = eval_call(code, mem)

                case "args":
                    res = eval_args(code, mem)

                case "oper" | "id":
                    res = eval_oper(code, mem)

        case AOT():
            res = eval_token(code, mem)
            # print(res)
            res = res,
    return res



from typing import Any
from abc import ABC, abstractmethod
from functools import reduce
from hhat_lang.interpreter.memory import Mem
from hhat_lang.datatypes.builtin_datatype import (
    builtin_array_types_dict
)
from hhat_lang.datatypes.base_datatype import DataType, DataTypeArray
from hhat_lang.utils.utils import get_types_set
from hhat_lang.interpreter.memory import Var


class MetaTypeFn(type):
    def __repr__(cls) -> str:
        return f"{cls.__name__}"


class MetaFn(ABC):
    token = "meta-default"

    def __init__(self, mem: Mem, *values: Any):
        self.mem = mem
        self.values = self.check_data(values)

    def check_data(self, data: Any) -> Any:
        if isinstance(data, tuple):
            if len(data) > 1:
                return tuple(self.check_data(k) for k in data)
            return self.check_data(data[0])
        if isinstance(data, (DataType, DataTypeArray)):
            return data
        if isinstance(data, Var):
            return self.mem.get_var(data.name).data

    @abstractmethod
    def __call__(self, values: Any | None = None) -> tuple[Any]:
        ...

    def __repr__(self):
        return self.token


class Sum(MetaFn):
    token = "sum"

    def __init__(self, mem: Mem, *values: Any):
        super().__init__(mem, *values)

    def __call__(self, *values: Any) -> tuple[Any]:
        types_set_self = get_types_set(*self.values)
        if len(types_set_self) == 1:
            if len(values) == 0:
                return reduce(lambda x, y: x + y, self.values),
            values = self.check_data(values)
            if len(values) == len(self.values):
                return (values * self.values),
            types_set_other = get_types_set(values)
            type_val_other = types_set_other.pop()
            other_res = reduce(lambda x, y: x + y, values)
            self_oper = map(lambda x: x + other_res, self.values)
            return builtin_array_types_dict[type_val_other](*self_oper),
        raise NotImplementedError("operation with more than one data type not implemented.")


class Times(MetaFn):
    token = "times"

    def __init__(self, mem: Mem, *values: Any):
        super().__init__(mem, *values)

    def __call__(self, *values: Any) -> tuple[Any]:
        types_set_self = get_types_set(*self.values)
        if len(types_set_self) == 1:
            if len(values) == 0:
                return reduce(lambda x, y: x * y, self.values),
            values = self.check_data(values)
            if len(values) == len(self.values):
                return (values * self.values),
            types_set_other = get_types_set(values)
            type_val_other = types_set_other.pop()
            other_res = reduce(lambda x, y: x * y, values)
            self_oper = map(lambda x: x * other_res, self.values)
            return builtin_array_types_dict[type_val_other](*self_oper),
        raise NotImplementedError("operation with more than one data type not implemented.")


class Print(MetaFn):
    token = "print"

    def __init__(self, mem: Mem, *values: Any):
        super().__init__(mem, *values)

    def __call__(self, *values: Any) -> tuple[Any]:
        if len(values) == 0:
            print(*tuple(self.check_data(k) for k in self.values))
        else:
            print(*tuple(self.check_data(k) for k in (values + self.values)))
        return self.values,


class QInit(MetaFn):
    token = "@init"

    def __init__(self, mem: Mem, *values: Any):
        super().__init__(mem, *values)

    def __call__(self, *values: Any) -> tuple[Any]:
        pass


builtin_fn_dict = {
    "sum": Sum,
    "times": Times,
    "print": Print
}
builtin_fn_list = tuple(builtin_fn_dict.keys())



from hhat_ast import AST


class Data:
    def __init__(self, data: AST):
        self.value = data.value
        self.type = data.type
        self.ast_type = data.ast_type
        # self.args = data.args
        # if len(self.args) > 0:
        #     self._iter = self.args
        # else:
        #     self._iter = self.value,

    # def __iter__(self):
    #     yield from self._iter

    def __repr__(self):
        if self.value:
            value = str(self.value)
            # args = f"[{len(self.args)}]" if len(self.args) > 0 else ""
        else:
            # value = "(" + " ".join(str(k) for k in self.args) + ")"
            value = "?"
            # args = ""
        return value  # + args


class PreEval:
    def __init__(self, parsed_code):
        self.code = parsed_code

    def run(self):
        return format_code(self.code)


def format_code(code):
    if isinstance(code, AST):
        if len(code) > 0:
            if code.ast_type in ["program", "expr", "array"]:
                return format_code(code.args),
            elif code.ast_type == "oper":
                return ((Data(code),) + format_code(code.args),)
            else:
                raise ValueError(f"{code.ast_type=} invalid.")
        else:
            return Data(code),
    if isinstance(code, tuple):
        res = ()
        for k in code:
            res += format_code(k)
        return res



from hhat_lang.syntax_trees.ast import (
    Main,
    Expr,
    ManyExpr,
    Operation,
    Array,
    Id,
    Literal
)
from arpeggio import visit_parse_tree, PTNodeVisitor
from arpeggio.cleanpeg import ParserPEG


class CST(PTNodeVisitor):
    def __init__(self, defaults=True, **kwargs):
        super().__init__(defaults=defaults, **kwargs)

    def visit_program(self, n, k):
        # return Program(*k)
        # return NewMain(*k)
        res = tuple(ManyExpr(p) for p in k)
        return Main(*res)

    def visit_exprs(self, n, k):
        # return Expr(*k)
        return Expr(*k)

    def visit_expr(self, n, k):
        if len(k) > 1:
            # return Expr(*k)
            return ManyExpr(*k)  # NewExpr(*k)
        return k[0]

    def visit_operation(self, n, k):
        if len(k) > 1:
            # return Operation(k[0], *k[1:])
            return Operation(k[0], *k[1:])
        # return Operation(k[0])
        return Operation(k[0])

    def visit_array(self, n, k):
        # return Array(*k)
        return Array(*k)

    def visit_id(self, n, k):
        # return Id(n.value)
        return Id(n.value)

    def visit_literal(self, n, k):
        return k[0]

    def visit_BOOL(self, n, k):
        # return Literal(n.value, "bool")
        return Literal(n.value, "bool")

    def visit_INT(self, n, k):
        # return Literal(n.value, "int")
        return Literal(n.value, "int")


def parse_code(code):
    peg_grammar = open("../grammar/grammar.peg", "r").read()
    parsed_code = ParserPEG(peg_grammar, "program", reduce_tree=False)
    pt = parsed_code.parse(code)
    return visit_parse_tree(pt, CST())



from abc import ABC, abstractmethod
from typing import Any, Iterable


class DataType(ABC):
    def __init__(self, value: Any):
        self.value = value
        self.data = self.cast()
        self.value = str(value)

    @property
    @abstractmethod
    def token(self):
        ...

    @property
    @abstractmethod
    def type(self):
        ...

    @abstractmethod
    def cast(self):
        ...

    @abstractmethod
    def __add__(self, other: Any) -> Any:
        ...

    @abstractmethod
    def __radd__(self, other: Any) -> Any:
        ...

    @abstractmethod
    def __mul__(self, other: Any) -> Any:
        ...

    @abstractmethod
    def __rmul__(self, other: Any) -> Any:
        ...

    def __len__(self) -> int:
        return 1

    def __iter__(self) -> Iterable:
        yield from (self,)

    def __repr__(self) -> str:
        return f"{self.data}"


class DataTypeArray(ABC):
    def __init__(self, *values: Any):
        self.value = values
        self.data = self.cast()

    @property
    @abstractmethod
    def token(self):
        ...

    @property
    @abstractmethod
    def type(self):
        ...

    @abstractmethod
    def cast(self) -> Any:
        ...

    @abstractmethod
    def __add__(self, other: Any) -> Any:
        ...

    @abstractmethod
    def __radd__(self, other: Any) -> Any:
        ...

    @abstractmethod
    def __mul__(self, other: Any) -> Any:
        ...

    @abstractmethod
    def __rmul__(self, other: Any) -> Any:
        ...

    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self) -> Iterable:
        yield from self.data

    def __repr__(self):
        return f"[{' '.join(str(k) for k in self.data)}]"



from copy import deepcopy
from typing import Any, Callable, Iterable
from uuid import uuid4, uuid3, NAMESPACE_OID
from dataclasses import dataclass, field
from hhat_lang.syntax_trees.ast import AOT, AST
from hhat_lang.datatypes.builtin_datatype import (
    builtin_array_types_dict
)
from hhat_lang.datatypes.base_datatype import DataType, DataTypeArray


def transform_token_type(data: AOT):
    from hhat_builtin_fn import builtin_fn_dict
    match data.type:
        case "bool":
            return True if data.token == "T" else False if data.token == "F" else None
        case "int":
            return int(data.token)
        case "oper" | "id":
            if data.token in builtin_fn_dict.keys():
                return builtin_fn_dict[data.token]
            return data.token


@dataclass(init=False)
class Data:
    # type: str
    value: tuple[Any]

    def __init__(self, *values: Any):
        self.value = values
        self.id = str(uuid3(NAMESPACE_OID, str(self.value)))

    def format_value(self, value: Any):
        if isinstance(value, tuple):
            return tuple(self.format_value(k) for k in value)
        if isinstance(value, AOT):
            return transform_token_type(value)
        if isinstance(value, AST):
            return (
                tuple(self.format_value(k) for k in value),
                self.format_value(value.node)
            )
        raise NotImplementedError(f"value of type {type(value)}, not implemented!")

    def opers(self, other: "Data", operation: Callable) -> "Data":
        if isinstance(other, Data):
            return Data(*tuple(map(operation, self.value, other.value)))
        raise ValueError(f"wrong type ({type(other)}).")

    def __add__(self, other: "Data") -> "Data":
        return self.opers(other, lambda x, y: x + y)

    def __radd__(self, other: "Data") -> "Data":
        return self.opers(other, lambda x, y: y + x)

    def __mul__(self, other: "Data") -> "Data":
        return self.opers(other, lambda x, y: x * y)

    def __rmul__(self, other: "Data") -> "Data":
        return self.opers(other, lambda x, y: y * x)

    def __str__(self):
        vals = " ".join(str(k) for k in self.value)
        if len(self) > 1:
            return f"({vals})"
        return vals

    def __len__(self) -> int:
        return len(self.value)

    def __iter__(self) -> Iterable:
        yield from self.value


class Var:
    token = "id"

    def __init__(self, name: str):
        self.initialized = False
        self.name = name if name else ""
        self.data = ()
        self.id = str(uuid4())
        self.type = ""

    def analyze_data(self, data: Any) -> Any:
        if isinstance(data, (DataType, DataTypeArray)):
            self.type = data.type
            return data
        if isinstance(data, tuple):
            types = set(k.type for k in data)
            if len(types) == 1 or (len(types) == 2 and "" in types):
                self.type = data[0].type
                return builtin_array_types_dict[self.type](*data)
            else:
                raise NotImplementedError("Var with more than one type not implemented.")
        if isinstance(data, (str, int, bool)):
            raise ValueError("got pure python data on variable. what to do?")
        raise ValueError(f"what is this? {type(data)} | {data}")

    def get_data(self) -> tuple:
        return self.data

    def get(self, item: Any = None) -> Any:
        if item is None:
            return self.get_data()
        return self[item]

    def __call__(self, *values: Any):
        if not self.initialized:
            self.data = self.analyze_data(values)
            self.initialized = True
            return self
        else:
            raise ValueError("Cannot set new values to initialized variable.")

    def __getitem__(self, item: Any) -> Any:
        return self.data[item]

    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self) -> Iterable:
        yield from self.data

    def __repr__(self) -> str:
        content = str(self.data)
        var_type = f"<{self.type}>"
        return "%" + self.name + var_type + (content if self.data else "")


class R:
    def __init__(
            self,
            ast_type: str,
            value: Any,
            is_concurrent: bool,
            role: str,
            execute_after: tuple | None
    ):
        self.type = ast_type
        self.value = value if isinstance(value, tuple) else (value,)
        self.id = str(uuid4())
        self.is_concurrent = is_concurrent
        self.role = role
        self.execute_after = execute_after if execute_after else ()

    def __hash__(self):
        return hash(self.value)

    def __len__(self) -> int:
        return len(self.value)

    def __iter__(self) -> Iterable:
        yield from self.value

    def __repr__(self) -> str:
        return self.type + "(" + " ".join(str(k) for k in self.value) + ")"


class Fn:
    def __init__(self, name: str, args: R | None, body: R):
        self.id = str(uuid4())
        self.name = name
        self.args = args
        self.body = body

    def __repr__(self) -> str:
        args = "(" + " ".join(str(p) for p in self.args) + ")" if self.args is not None else ""
        return f"fn%{self.name}{args}({self.body})"


@dataclass
class Mem:
    parent_id: str = ""
    id: str = field(init=False, default=str(uuid4()))
    data: dict = field(init=False, default_factory=dict)

    def __post_init__(self):
        self.data = self._reset_data()

    @staticmethod
    def _reset_data() -> dict:
        return dict(
            shared=dict(
                stack=(),
                data=(),
                exprs=(),
                vars=dict(),
                fn=dict(),
                main=dict()
            ),
            pvt=dict(
                stack=(),
                data=(),
                exprs=(),
            )
        )

    @property
    def var_list(self) -> tuple:
        return tuple(self.data["shared"]["data"].keys())

    @property
    def fn_list(self) -> tuple:
        return tuple(self.data["shared"]["fn"].keys())

    def pop_stack(self, key: str = "shared") -> Any:
        res = self.data[key]["stack"][-1]
        self.data[key]["stack"] = tuple(self.data[key]["stack"][:-1])
        return res

    def pop_expr(self, key: str = "shared") -> Any:
        res = self.data[key]["exprs"][-1]
        self.data[key]["exprs"] = tuple(self.data[key]["exprs"][:-1])
        return res

    def get_stack(self, key: str = "shared") -> tuple[Any]:
        res = deepcopy(self.data[key]["stack"])
        self.data[key]["stack"] = ()
        return res

    def get_data(self, key: str = "shared") -> tuple[Any]:
        res = deepcopy(self.data[key]["data"])
        self.data[key]["data"] = ()
        return res

    def get_var(self, name: str, key: str = "shared") -> Any:
        return self.data[key]["vars"][name]["data"]

    def get_expr(self, key: str = "shared") -> tuple[Any]:
        expr = self.data[key]["exprs"][-1]
        self.data[key]["exprs"] = tuple(self.data[key]["exprs"][:-1])
        return expr

    def get_fn(self, name: str, n_args: int, args: Any) -> Any:
        # TODO: implement it properly
        # return self.data["shared"]["fn"][n_args]
        raise NotImplemented("Mem.get_fn not implemented yet.")

    def put_stack(self, value: Any, key: str = "shared") -> None:
        self.data[key]["stack"] += value,

    def put_expr(self, value: Any, key: str = "shared") -> None:
        # self.data[key]["exprs"][value.id] = dict(data=value)
        # return value.id
        self.data[key]["exprs"] += value,

    def put_data(self, value: Data | DataType | DataTypeArray, key: str = "shared") -> None:
        self.data[key]["data"] += value,

    def put_var(self, data: Var, scope_id: str, key: str = "shared") -> tuple[str, str]:
        self.data[key]["vars"][data.name] = dict(data=data, scope_id=scope_id)
        return data.id, scope_id

    def put_fn(self, data: Fn):
        # TODO: implement properly how to deal with the data args
        mem_fn = self.data["shared"]["fn"]
        if data.name not in mem_fn:
            mem_fn[data.name] = {len(data.args): {data.args: data.body}}
        else:
            if len(data.args) in mem_fn[data.name]:
                mem_fn[data.name][len(data.args)].update({data.args: data.body})
            else:
                mem_fn[data.name][len(data.args)] = {data.args: data.body}

    def share_stack(self, mem_target: "Mem") -> None:
        mem_target.data["shared"]["stack"] += self.data["shared"]["stack"]

    def share_vars(self, mem_target: "Mem") -> None:
        mem_target.data["shared"]["vars"].update(self.data["shared"]["vars"])

    def share_data(self, mem_target: "Mem") -> None:
        mem_target.data["shared"]["data"] += self.data["shared"]["data"]

    def clear_stack(self, key: str = "shared") -> None:
        self.data[key]["stack"] = ()

    def clear_data(self, key: str = "shared") -> None:
        self.data[key]["data"] = ()

    def clear_var(self, key: str = "shared") -> None:
        self.data[key]["vars"] = dict()

    def clear_exprs(self, key: str = "shared") -> None:
        self.data[key]["exprs"] = ()

    def clear_all(self) -> None:
        self.data = self._reset_data()

    def __contains__(self, item: Any) -> bool:
        return (
            item in self.data["shared"]["stack"]
            or item in self.data["shared"]["data"]
            or item in self.data["shared"]["vars"].keys()
            or item in self.data["shared"]["exprs"]
        )



from setuptools import setup
import pathlib


here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")
requirements = open("requirements.txt", "r").readlines()
version = open("hhat_lang/version.txt", "r").readline()


setup(
    name="hhat-lang",
    version=version,
    description="H-hat: a high level abstraction quantum programming language.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hhat-lang/hhat_lang",
    author="Eduardo Maschio (Doomsk)",
    author_email="eduardo.maschio@hhat-lang.org",
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development",
        "Topic :: Software Development :: Interpreters",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    keywords="programming language, quantum",
    python_requires=">=3.10, <4",
    install_requires=requirements,
    include_package_data=True,
    extra_require={
        # "netqasm": ["netqasm"],
    },
    entry_points={
        "console_scripts": ["hhat=hhat_lang.exec:main"]
    },
    project_urls={"Unitary Fund": "https://unitary.fund/"}
)



from hhat_lang.interpreter import parse_code, Analysis, Eval, R
from hhat_lang.syntax_trees import AST
from hhat_lang import __version__
import click


def read_file(file: str) -> str:
    if file.endswith(".hat"):
        return open(file, "r").read()


def execute_parsing_code(c: str, verbose: bool = False) -> AST:
    pc_ = parse_code(c)
    if verbose:
        print(f"- code:\n{c}\n")
        print(f"- parsed code:\n{pc_}\n")
    return pc_


def execute_analysis(pc: AST, verbose: bool = False) -> R:
    if verbose:
        print(f"- analysis (pre-evaluation):")
    analysis = Analysis(pc)
    res_ = analysis.run()
    print("\n")
    return res_


def execute_eval(c: R) -> None:
    ev_ = Eval(c)
    print("- executing code:\n")
    ev_.run()


def run_codes(c: str, verbose: bool = False) -> None:
    pc_ = execute_parsing_code(c, verbose)
    pev_ = execute_analysis(pc_, verbose)
    print("-" * 80)
    execute_eval(pev_)


@click.group(invoke_without_command=True, context_settings=dict(ignore_unknown_options=True))
@click.argument("file", type=click.Path(exists=True), required=False)
@click.option("-v", "--version", "version", is_flag=True)
@click.option("--verbose", "verbose", is_flag=True)
def main(file, version, verbose):
    if version:
        click.echo(f"H-hat version {__version__}")
    else:
        if file:
            code_text = read_file(file)
            run_codes(code_text, verbose=verbose)
        else:
            # TODO: make a REPL?
            pass



import re


def literal_bool_define(value: str) -> bool:
    bool_vals = dict(T=True, F=False)
    if res := bool_vals.get(value, False):
        return res
    raise ValueError(f"wrong value for boolean ({value}).")


def literal_int_define(value: str) -> int:
    if res := re.match(r"^(0|-?[1-9][0-9]*)$", value):
        return int(res.group())
    raise ValueError(f"wrong value for integer ({value}).")
